/*----------------------------------------------------------------------------*-

  ttrd2-08a-t0405a-v001a_processor.c (Release 2018-06-07a)

  Configures Processor in one of the following Modes / States after a reset:

    FAIL_SAFE_S
    NORMAL_01_M
    NORMAL_02_M
    NORMAL_03_M

  Implements rudimentary Fail-Safe State. 

  Targets DuplicaTTor Evaluation Board (STM32F405, v002).

  Code can be loaded onto either MCU.

-*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*-

  This code is copyright (c) 2014-2018 SafeTTy Systems Ltd.

  This code forms part of a Time-Triggered Reference Design (TTRD) that is 
  documented in the following book: 

   Pont, M.J. (2016) 
   "The Engineering of Reliable Embedded Systems (Second Edition)", 
   Published by SafeTTy Systems Ltd. ISBN: 978-0-9930355-3-1.

  Both the TTRDs and the above book ("ERES2") describe patented 
  technology and are subject to copyright and other restrictions.

  This code may be used without charge: [i] by universities and colleges on 
  courses for which a degree up to and including 'MSc' level (or equivalent) 
  is awarded; [ii] for non-commercial projects carried out by individuals 
  and hobbyists.

  Any and all other use of this code and / or the patented technology 
  described in ERES2 requires purchase of a ReliabiliTTy Technology Licence:
  http://www.safetty.net/technology/reliabilitty-technology-licences

  Please contact SafeTTy Systems Ltd if you require clarification of these 
  licensing arrangements: http://www.safetty.net/contact

  CorrelaTTor, DecomposiTTor, DuplicaTTor, MoniTTor, PredicTTor, ReliabiliTTy,  
  SafeTTy, SafeTTy Systems, TriplicaTTor and WarranTTor are registered 
  trademarks or trademarks of SafeTTy Systems Ltd in the UK & other countries.

-*----------------------------------------------------------------------------*/

// Processor Header
#include "..\main\main.h"

// Tasks
#include "..\tasks\ttrd2-08d-t0405a-v001a_heartbeat_amber_task.h"
#include "..\tasks\ttrd2-08d-t0405a-v001a_heartbeat_green_task.h"
#include "..\tasks\ttrd2-08d-t0405a-v001a_heartbeat_red_task.h"
#include "..\tasks\ttrd2-08d-t0405a-v001a_iwdt_task.h"
#include "..\tasks\ttrd2-08d-t0405a-v001a_processor_task.h"
#include "..\tasks\ttrd2-08d-t0405a-v001a_switch_task.h"

// ------ Private Duplicated Variables ('File Global') -----------------------

// The current system Mode / State
static eProc_MoSt Processor_MoSt_g;

// ------ Private function prototypes (declarations) -------------------------

static void       PROCESSOR_Identify_Reqd_MoSt(void);
static void       PROCESSOR_Configure_Reqd_MoSt(void);

/*----------------------------------------------------------------------------*-

  PROCESSOR_Init()

  Wrapper for system startup functions.
  
  PARAMETERS:
     None.

  LONG-TERM DATA:
     None.

  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void PROCESSOR_Init(void)
   {
   PROCESSOR_Identify_Reqd_MoSt();
   PROCESSOR_Configure_Reqd_MoSt();
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Identify_Reqd_MoSt()

  Identify the cause of the system reset and set the system mode accordingly.

  PARAMETERS:
     None.

  LONG-TERM DATA:
     Processor_MoSt_g (W)

  MCU HARDWARE:
     Registers indicating cause of reset (R).

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void PROCESSOR_Identify_Reqd_MoSt(void)
   {
   if (RCC_GetFlagStatus(RCC_FLAG_IWDGRST))
      {
      // WDT reset: assume a fault.
      PROCESSOR_Set_MoSt(FAIL_SAFE_S);
      }
   else if (RCC_GetFlagStatus(RCC_FLAG_SFTRST))
      {
      // Software reset (note - may include debugger)
      // Not used for Mode changes here: assume a fault
      PROCESSOR_Set_MoSt(FAIL_SAFE_S);
      }
   else if (RCC_GetFlagStatus(RCC_FLAG_PORRST))
      {
      // Power-On Reset
      // Start in 'NORMAL_01_M' mode
      PROCESSOR_Set_MoSt(NORMAL_01_M);
      }
   else if (RCC_GetFlagStatus(RCC_FLAG_PINRST))
      {
      // 'Reset button' pressed on DEB-0405
      // Here we treat this is the same way as a POR
      PROCESSOR_Set_MoSt(NORMAL_01_M);
      }
   else
      {
      // Other possible reset causes
      // Low-power management reset / Windowed WDT / Brown-out
      // In all cases we enter fail-safe state
      PROCESSOR_Set_MoSt(FAIL_SAFE_S);
      }   
  
   // Clear all reset flags
   RCC_ClearFlag();
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Configure_Reqd_MoSt()

  Configure the system in the required mode.  

  PARAMETERS:
     None.

  LONG-TERM DATA:
     System_node_G (R)

  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.
   
-*----------------------------------------------------------------------------*/
void PROCESSOR_Configure_Reqd_MoSt(void)
   {
   switch (Processor_MoSt_g)
      {
      default:          // Default to "FAIL_SAFE_S"
      case FAIL_SAFE_S:
         {
         // Trigger "fail safe" behaviour
         PROCESSOR_Perform_Safe_Shutdown();

         break;
         }

      case NORMAL_01_M:
         {
         // Set up scheduler for 5 ms ticks
         SCH_Init_Milliseconds(5);

         // Set up WDT 
         // Timeout is parameter * 125 s: param 100 => ~12.5 ms
         // NOTE: WDT driven by RC oscillator - timing varies with temperature            
         WATCHDOG_Init(100);

         // Prepare for heartbeat task (red)
         HEARTBEAT_RED_Init();

         // Prepare for the switch-reading task
         SWITCH_BTN_Init(); 
         
         // Prepare for the Processor Task
         // Controls Mode / State changes
         // Checks CPU temperature
         PROCESSOR_TASK_Init();

         // Add tasks to schedule.
         // Parameters are:
         // A. Task name
         // B. Initial delay / offset (in Ticks)
         // C. Task period (in Ticks): Must be > 0
         // D. Name of task deinit function
         //
         //           A                       B   C    D    
         SCH_Add_Task(WATCHDOG_Update,        0,  1,   WATCHDOG_Deinit);     
         SCH_Add_Task(HEARTBEAT_RED_Update,   0,  200, HEARTBEAT_RED_Deinit);   
         SCH_Add_Task(SWITCH_BTN_Update,      10, 2,   SWITCH_BTN_Deinit);   
         SCH_Add_Task(PROCESSOR_TASK_Update,  0,  200, PROCESSOR_TASK_Deinit);   

         // Feed the watchdog
         WATCHDOG_Update();

         break;
         }

      case NORMAL_02_M:
         {
         // Set up scheduler for 5 ms ticks
         SCH_Init_Milliseconds(5);

         // Set up WDT 
         // Timeout is parameter * 125 s: param 100 => ~12.5 ms
         // NOTE: WDT driven by RC oscillator - timing varies with temperature            
         WATCHDOG_Init(100);

         // Prepare for heartbeat task (amber)
         HEARTBEAT_AMBER_Init();

         // Prepare for the switch-reading task
         SWITCH_BTN_Init(); 
         
         // Prepare for the Processor Task
         // Controls Mode / State changes
         // Checks CPU temperature
         PROCESSOR_TASK_Init();

         // Add tasks to schedule.
         // Parameters are:
         // A. Task name
         // B. Initial delay / offset (in Ticks)
         // C. Task period (in Ticks): Must be > 0
         // D. Name of task deinit function
         //
         //           A                       B   C    D    
         SCH_Add_Task(WATCHDOG_Update,        0,  1,   WATCHDOG_Deinit);     
         SCH_Add_Task(HEARTBEAT_AMBER_Update, 0,  200, HEARTBEAT_AMBER_Deinit);   
         SCH_Add_Task(SWITCH_BTN_Update,      10, 2,   SWITCH_BTN_Deinit);   
         SCH_Add_Task(PROCESSOR_TASK_Update,  0,  200, PROCESSOR_TASK_Deinit);   

         // Feed the watchdog
         WATCHDOG_Update();

         break;
         }

      case NORMAL_03_M:
         {
         // Set up scheduler for 5 ms ticks
         SCH_Init_Milliseconds(5);

         // Set up WDT 
         // Timeout is parameter * 125 s: param 100 => ~12.5 ms
         // NOTE: WDT driven by RC oscillator - timing varies with temperature            
         WATCHDOG_Init(100);

         // Prepare for heartbeat task (green)
         HEARTBEAT_GREEN_Init();

         // Prepare for the switch-reading task
         SWITCH_BTN_Init(); 
         
         // Prepare for the Processor Task
         // Controls Mode / State changes
         // Checks CPU temperature
         PROCESSOR_TASK_Init();

         // Add tasks to schedule.
         // Parameters are:
         // A. Task name
         // B. Initial delay / offset (in Ticks)
         // C. Task period (in Ticks): Must be > 0
         // D. Name of task deinit function
         //
         //           A                       B   C    D    
         SCH_Add_Task(WATCHDOG_Update,        0,  1,   WATCHDOG_Deinit);     
         SCH_Add_Task(HEARTBEAT_GREEN_Update, 0,  200, HEARTBEAT_GREEN_Deinit);   
         SCH_Add_Task(SWITCH_BTN_Update,      10, 2,   SWITCH_BTN_Deinit);   
         SCH_Add_Task(PROCESSOR_TASK_Update,  0,  200, PROCESSOR_TASK_Deinit);   

         // Feed the watchdog
         WATCHDOG_Update();

         break;
         }
      }
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Set_MoSt()

  Set the value of 'Processor_MoSt_g' (Duplicated Variable). 
   
  PARAMETERS:
     MOST : Value to be assigned.

  LONG-TERM DATA:
     Processor_MoSt_g (W)
     Processor_MoSt_ig (W)
   
  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void PROCESSOR_Set_MoSt(const eProc_MoSt MOST)
   {
   Processor_MoSt_g = MOST;
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Get_MoSt()

  Returns the value of 'Processor_MoSt_g' (Duplicated Variable). 
   
  PARAMETERS:
     None.

  LONG-TERM DATA:
     Processor_MoSt_g (R)
     Processor_MoSt_ig (R)
   
  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     The value of 'Processor_MoSt_g' (Duplicated Variable). 

-*----------------------------------------------------------------------------*/
eProc_MoSt PROCESSOR_Get_MoSt(void)
   {
   return Processor_MoSt_g;
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Change_MoSt()

  Force software reset, which will (at minimum) cause a 'Same-Mode Reset'.

  Note.  Before calling this function, PROCESSOR_Store_Reqd_MoSt() will 
  usually be called (to indicate the required new Mode / State). 

  PARAMETERS:
     None.

  LONG-TERM DATA:
     None.
       
  MCU HARDWARE:
      NVIC.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void PROCESSOR_Change_MoSt(void)
   {
   // NOTE: Required MoSt needs to be set before this function is called

   // Feed the watchdog
   WATCHDOG_Update();
 
   // Disable scheduler (and delete all tasks)
   SCH_Deinit();

   // Feed the watchdog
   WATCHDOG_Update();

   // Add required Tasks to new schedule
   PROCESSOR_Configure_Reqd_MoSt(); 

   // Restart the scheduler
   SCH_Start();
   }

/*----------------------------------------------------------------------------*-

  PROCESSOR_Perform_Safe_Shutdown()

  Basic attempt to move the system into a safe ('silent') state.

  PARAMETERS:
     None.

  LONG-TERM DATA:
     None.

  MCU HARDWARE:
     Heartbeat LED (GPIO pin).

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.
   
-*----------------------------------------------------------------------------*/
void PROCESSOR_Perform_Safe_Shutdown(void)
   {
   static uint32_t State_s = 0;
   uint32_t Delay1, Delay2;

   // Here we simply "fail silent" with rudimentary fault reporting.
   // OTHER BEHAVIOUR IS LIKELY TO BE REQUIRED IN YOUR DESIGN

   // *************************************
   // NOTE: This function should NOT return
   // *************************************

   // Set up Heartbeat LED pin
   HEARTBEAT_AMBER_Init();
   HEARTBEAT_GREEN_Init();
   HEARTBEAT_RED_Init();

   while(1)
      {
      // Flash all LEDs to indicate fault
      for (Delay1 = 0; Delay1 < 20000000; Delay1++) 
         {
         Delay2 *= 3;
         }

      if (State_s == 0)
         {
         GPIO_SetBits(RED_LED_PORT, RED_LED_PIN);
         GPIO_SetBits(AMBER_LED_PORT, AMBER_LED_PIN);
         GPIO_SetBits(GREEN_LED_PORT, GREEN_LED_PIN);
         State_s = 1;
         }
      else
         {
         GPIO_ResetBits(RED_LED_PORT, RED_LED_PIN);
         GPIO_ResetBits(AMBER_LED_PORT, AMBER_LED_PIN);
         GPIO_ResetBits(GREEN_LED_PORT, GREEN_LED_PIN);
         State_s = 0;
         }
      }
   }

/*----------------------------------------------------------------------------*-
  ------------------------------ END OF FILE ---------------------------------
-*----------------------------------------------------------------------------*/
