/*----------------------------------------------------------------------------*-

  ttrd2-08d-t0405a-v001a_iwdt.c (Release 2018-06-07a)

  ----------------------------------------------------------------------------
   
  'Watchdog' timer library (internal).

  Targets DuplicaTTor Evaluation Board (STM32F405, v002).

  Code can be loaded onto either MCU.

-*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*-

  This code is copyright (c) 2014-2018 SafeTTy Systems Ltd.

  This code forms part of a Time-Triggered Reference Design (TTRD) that is 
  documented in the following book: 

   Pont, M.J. (2016) 
   "The Engineering of Reliable Embedded Systems (Second Edition)", 
   Published by SafeTTy Systems Ltd. ISBN: 978-0-9930355-3-1.

  Both the TTRDs and the above book ("ERES2") describe patented 
  technology and are subject to copyright and other restrictions.

  This code may be used without charge: [i] by universities and colleges on 
  courses for which a degree up to and including 'MSc' level (or equivalent) 
  is awarded; [ii] for non-commercial projects carried out by individuals 
  and hobbyists.

  Any and all other use of this code and / or the patented technology 
  described in ERES2 requires purchase of a ReliabiliTTy Technology Licence:
  http://www.safetty.net/technology/reliabilitty-technology-licences

  Please contact SafeTTy Systems Ltd if you require clarification of these 
  licensing arrangements: http://www.safetty.net/contact

  CorrelaTTor, DecomposiTTor, DuplicaTTor, MoniTTor, PredicTTor, ReliabiliTTy,  
  SafeTTy, SafeTTy Systems, TriplicaTTor and WarranTTor are registered 
  trademarks or trademarks of SafeTTy Systems Ltd in the UK & other countries.

-*----------------------------------------------------------------------------*/

// Module header
#include "ttrd2-08d-t0405a-v001a_iwdt_task.h"

// ------ Private variables ('File Global') -----------------------

static uint32_t Watchdog_running_g = WDT_NOT_RUNNING;

/*----------------------------------------------------------------------------*-

  WATCHDOG_Init()

  Set up watchdog timer.
 
  The watchdog timer is driven by the Internal LSI (RC) Oscillator:
  the timing varies (significantly) with temperature.

  Datasheet values: 
  RC oscillator frequency (Minimum) : 17 kHz  
  RC oscillator frequency (Minimum) : 32 kHz 
  RC oscillator frequency (Minimum) : 47 kHz 

  It is suggested that you use twice the required timeout value (approx).

  *** NOTES ***

  Does *not* start the WDT (WDT starts first time the Task is released).

  Does not configure the watchdog if it is already running, to simplify the
  process of performing manual (reset free) Mode changes.
    
  PARAMETERS:
     WDT_COUNT : Will be multiplied by ~125µs to determine the timeout. 
   
  LONG-TERM DATA:
     None.

  MCU HARDWARE:
     iWDT.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void WATCHDOG_Init(const uint32_t WDT_COUNT)
   {
   // Don't try to start the WDT if it is already running
   if (Watchdog_running_g == WDT_NOT_RUNNING)
      {
      // Enable write access to IWDG_PR and IWDG_RLR registers
      IWDG->KR = 0x5555;

      // Set pre-scalar to 4 (timer resolution is 125µs)
      IWDG->PR = 0x00;

      // Counts down to 0 in increments of 125µs
      // Max reload value is 0xFFF (4095) or ~511 ms (with this prescalar)
      IWDG->RLR = WDT_COUNT;

      // Reload IWDG counter
      IWDG->KR = 0xAAAA;
      }
   else
      {
      // If the WDT is running then feed it
      WATCHDOG_Update();
      }
   }


/*----------------------------------------------------------------------------*-

  WATCHDOG_Deinit()

  Deinit function for WDT.
    
  PARAMETERS:
     None. 
   
  LONG-TERM DATA:
     None.

  MCU HARDWARE:
     iWDT.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  RETURN VALUE:
     None.

-*----------------------------------------------------------------------------*/
void WATCHDOG_Deinit(void)
   {
   // Nothing to do (cannot stop WDT)
   }

/*----------------------------------------------------------------------------*-

  WATCHDOG_Update()

  Feed the watchdog timer.

  See Watchdog_Init() for further information.
   
  PARAMETERS:
     None.

  LONG-TERM DATA (all DuVs):
     None.
   
  MCU HARDWARE:
     None.

  PRE-CONDITION CHECKS:
     None.

  POST-CONDITION CHECKS:
     None.

  ERROR DETECTION / ERROR HANDLING:
     None.

  WCET:
     Not yet determined.

  BCET:
     Not yet determined.

  RETURN VALUE:
     RETURN_NORMAL_STATE (fixed).

-*----------------------------------------------------------------------------*/
void WATCHDOG_Update(void)
   {
   static uint32_t First_call_s = TRUE;  

   if (First_call_s == TRUE)
      {
      First_call_s = FALSE;

      // Enable IWDG (the LSI oscillator will be enabled by hardware)
      IWDG->KR = 0xCCCC;

      // Store the state
      Watchdog_running_g = WDT_RUNNING;
      }

   // Feed the watchdog (reload IWDG counter)
   IWDG->KR = 0xAAAA;
   }

/*----------------------------------------------------------------------------*-
  ------------------------------ END OF FILE ---------------------------------
-*----------------------------------------------------------------------------*/
